#
# This file was generated using xslt from its XML file
#
# Copyright 2008, Associated Universities Inc., Washington DC
#
import sys
import os
from casac import *
import string
import time
import inspect
import gc
import numpy
from odict import odict
from task_simalma import simalma
from task_simalma import casalog

class simalma_pg_:
    __name__ = "simalma"

    def __init__(self) :
       self.__bases__ = (simalma_pg_,)
       self.__doc__ = self.__call__.__doc__


    def __call__(self, project=None,dryrun=None,skymodel=None,inbright=None,indirection=None,incell=None,incenter=None,inwidth=None,complist=None,compwidth=None,setpointings=None,ptgfile=None,integration=None,direction=None,mapsize=None,antennalist=None,hourangle=None,totaltime=None,tpnant=None,tptime=None,pwv=None,image=None,imsize=None,imdirection=None,cell=None,niter=None,threshold=None,graphics=None,verbose=None,overwrite=None,):

        """Simulation task for ALMA   
    -------------------------------
    Parameters:

    project -- root filename for all output files.  A subdirectory will be 
         created, and all created files will be placed in that subdirectory
         including the informational report.
    -------------------------------
    skymodel -- input image (used as a model of the sky)
       * simalma requires a CASA or fits image. If you merely have a grid of 
         numbers, you will need to write them out as fits or write a 
         CASA script to read them in and use the ia tool to create an image 
         and insert the data.
       
       * simalma does NOT require a coordinate system in the header. If the
         coordinate information is incomplete, missing, or you would like to
         override it, set the appropriate "in" parameters. NOTE that setting 
         those parameters simply changes the header values, ignoring
         any values already in the image. No regridding is performed.
   
       * If you have a proper Coordinate System, simalma will do its best to
         generate visibilities from that, and then create a synthesis image
         according to the specified user parameters. 

       * You can manipulate an image header manually with the "imhead" task.

    inbright -- peak brightness in Jy/pixel, or "" for unchanged
       * NOTE: "unchanged" will take the numerical values in your image 
         and assume they are in Jy/pixel, even if it says some other unit 
         in the header. 
    indirection -- central direction, or "" for unchanged
    incell -- spatial pixel size, or "" for unchanged
    incenter -- frequency of center channel e.g. "89GHz", or "" for unchanged
    inwidth -- width of channels, or "" for unchanged - this should be a 
         string representing a quantity with units e.g. "10MHz"
       * NOTE: only works reliably with frequencies, not velocities
       * NOTE: it is not possible to change the number of spectral planes
         of the sky model, only to relabel them with different frequencies
         That kind of regridding can be accomplished with the CASA toolkit.

    -------------------------------
    complist -- component list model of the sky, added to or instead of skymodel
    compwidth -- bandwidth of components; if simulating from components only, 
         this defines the bandwidth of the MS and output images

    -------------------------------
    setpointings -- calculate a map of pointings, or if false, the user should
         provide a ptgfile
       * if graphics are on, display the pointings shown on the model image
       * if a list of directions is not specified, observations with the 
         ALMA 12m and ACA 7m arrays will observe a region of size "mapsize" 
         using the same hexagonal algorithm as the ALMA OT, with Nyquist 
         sampling.
       * The total power array maps a slightly (+1 primary beam) larger area 
         than the 12m array does, to improve later image combination. 
         It samples the region with lattice grids of spacing 0.33 lambda/D.

    ptgfile -- a text file specifying directions in the same 
         format as the example, and optional integration times, e.g.
         #Epoch     RA          DEC      TIME(optional)
         J2000 23h59m28.10 -019d52m12.35 10.0
       * if the time column is not present in the file, it will use
         "integration" for all pointings.
       * NOTE: at this time the file should contain only science pointings:
         simalma will observe these until totaltime is used up. 
    integration --- Time interval for each integration e.g '10s'
       * NOTE: to simulate a "scan" longer than one integration, use 
         setpointings to generate a pointing file, and then edit the 
         file to increase the time at each point to be larger than 
         the parameter integration time.
    direction -- mosaic center direction e.g 'J2000 19h00m00 -40d00m00'
       * can optionally be a list of pointings
       * otherwise simalma will pack mapsize with grids proper for the
         array (see below).
    mapsize -- angular size of map e.g. '40arcsec' or ['1arcmin','30arcsec']
       * set to "" to span the model image
    -------------------------------
    antennalist -- vector of ascii files containing antenna positions, 
         one for each configuration of 7m or 12m dishes.
       * NOTE: In this task, it should be an ALMA configuration.
       * standard arrays are found in your CASA data repository, 
         os.getenv("CASAPATH").split()[0]+"/data/alma/simmos/"        
       * a string of the form "alma;0.5arcsec" will be parsed into a 
         12m ALMA configuration - see casaguides.nrao.edu
       * examples: ['alma.cycle2.5.cfg','aca.cycle2.i.cfg']
              ['alma.cycle1;0.3arcsec','alma.cycle1.1.cfg','aca.i.cfg']
    hourangle -- hour angle of observation e.g. '-3h'
       * note that if you don't add a unit, it will assume seconds.
    totaltime --- total time of observations. This should either be a scalar
         time quantity expressed as a string e.g. '1h', '3600sec', '10min', 
         or a vector of such quantities, corresponding to the elements of 
         the antennalist vector, e.g. ['5min','20min','3h'].  If you 
         specify a scalar, that will be used for the highest resolution 
         12m configuration in antennalist, and any lower resolution 12m 
         configurations, any 7m configurations, and any TP configurations
         will have observing times relative to totaltime of 0.5, 2,and 4, 
         respectively.
    -------------------------------
    tpnant -- the number of total power antennas to use in simulation.
    tptime -- if tpnant>0, the user must specify the observing time for 
         total power as a CASA quantity e.g. '4h'.  
       * NOTE: in CASA 4.2 this is not broken up among multiple days - 
         a 20h track will include observations below the horizon, 
         which is probably not what is desired.
    -------------------------------
    pwv -- precipitable water vapor if constructing an atmospheric model.
         Set 0 for noise-free simulation. When pwv>0, thermal noise is
         applied to the simulated data.
       * J. Pardo's ATM library will be used to construct anatmospheric
         profile for the ALMA site:
         altitude 5000m, ground pressure 650mbar, relhum=20%, 
         a water layer of pwv at altitude of 2km,
         the sky brightness temperature returned by ATM, and internally
         tabulated receiver temperatures.
       See the documents of simobserve for more details.
    -------------------------------
    image -- option to invert and deconvolve the simulated measurement set(s)
       * NOTE: interactive clean or more parameters than the subset visible
         here are available by simply running the clean task directly.
       * if graphics turned on, display the clean image and residual image
       * uses Cotton-Schwab clean for single fields and Mosaic gridding
         for multiple fields (with Clark PSF calculation in minor cycles).
    imsize -- image size in spatial pixels (x,y)
       0 or -1 will use the model image size; example: imsize=[500,500]
    imdirection -- phase center for synthesized image.  default is to 
       center on the sky model.
    cell -- cell size e.g '10arcsec'.  "" defaults to the skymodel cell
    niter -- number of clean/deconvolution iterations, 0 for no cleaning
    threshold -- flux level to stop cleaning
    -------------------------------
    graphics -- view plots on the screen, saved to file, both, or neither
    verbose -- print extra information to the logger and terminal
    overwrite -- overwrite existing files in the project subdirectory
    -------------------------------

    Please see the documents of simobserve and simanalyze for
    the list of outputs produced.

  
        """
        a=inspect.stack()
        stacklevel=0
        for k in range(len(a)):
          if (string.find(a[k][1], 'ipython console') > 0) or (string.find(a[k][1], '<string>') >= 0):
                stacklevel=k
                break
        myf=sys._getframe(stacklevel).f_globals
        myf['__last_task'] = 'simalma'
        myf['taskname'] = 'simalma'
        ###
        myf['update_params'](func=myf['taskname'],printtext=False)
        ###
        ###
        #Handle globals or user over-ride of arguments
        #
	function_signature_defaults=dict(zip(self.__call__.func_code.co_varnames,self.__call__.func_defaults))
	useLocalDefaults = False

        for item in function_signature_defaults.iteritems():
                key,val = item
                keyVal = eval(key)
                if (keyVal == None):
                        #user hasn't set it - use global/default
                        pass
                else:
                        #user has set it - use over-ride
			if (key != 'self') :
			   useLocalDefaults = True
                        #myf[key]=keyVal

	myparams = {}
	if useLocalDefaults :
	   for item in function_signature_defaults.iteritems():
	       key,val = item
	       keyVal = eval(key)
	       exec('myparams[key] = keyVal')
	       if (keyVal == None):
	           exec('myparams[key] = '+ key + ' = self.itsdefault(key)')
		   keyVal = eval(key)
		   if(type(keyVal) == dict) :
		      exec('myparams[key] = ' + key + ' = keyVal[len(keyVal)-1][\'value\']')

	else :
            uselessvariable = None 
            myparams['project'] = project = myf['project']
            myparams['dryrun'] = dryrun = myf['dryrun']
            myparams['skymodel'] = skymodel = myf['skymodel']
            myparams['inbright'] = inbright = myf['inbright']
            myparams['indirection'] = indirection = myf['indirection']
            myparams['incell'] = incell = myf['incell']
            myparams['incenter'] = incenter = myf['incenter']
            myparams['inwidth'] = inwidth = myf['inwidth']
            myparams['complist'] = complist = myf['complist']
            myparams['compwidth'] = compwidth = myf['compwidth']
            myparams['setpointings'] = setpointings = myf['setpointings']
            myparams['ptgfile'] = ptgfile = myf['ptgfile']
            myparams['integration'] = integration = myf['integration']
            myparams['direction'] = direction = myf['direction']
            myparams['mapsize'] = mapsize = myf['mapsize']
            myparams['antennalist'] = antennalist = myf['antennalist']
            myparams['hourangle'] = hourangle = myf['hourangle']
            myparams['totaltime'] = totaltime = myf['totaltime']
            myparams['tpnant'] = tpnant = myf['tpnant']
            myparams['tptime'] = tptime = myf['tptime']
            myparams['pwv'] = pwv = myf['pwv']
            myparams['image'] = image = myf['image']
            myparams['imsize'] = imsize = myf['imsize']
            myparams['imdirection'] = imdirection = myf['imdirection']
            myparams['cell'] = cell = myf['cell']
            myparams['niter'] = niter = myf['niter']
            myparams['threshold'] = threshold = myf['threshold']
            myparams['graphics'] = graphics = myf['graphics']
            myparams['verbose'] = verbose = myf['verbose']
            myparams['overwrite'] = overwrite = myf['overwrite']

        if type(direction)==str: direction=[direction]
        if type(mapsize)==str: mapsize=[mapsize]
        if type(antennalist)==str: antennalist=[antennalist]
        if type(totaltime)==str: totaltime=[totaltime]
        if type(imsize)==int: imsize=[imsize]

	result = None

#
#    The following is work around to avoid a bug with current python translation
#
        mytmp = {}

        mytmp['project'] = project
        mytmp['dryrun'] = dryrun
        mytmp['skymodel'] = skymodel
        mytmp['inbright'] = inbright
        mytmp['indirection'] = indirection
        mytmp['incell'] = incell
        mytmp['incenter'] = incenter
        mytmp['inwidth'] = inwidth
        mytmp['complist'] = complist
        mytmp['compwidth'] = compwidth
        mytmp['setpointings'] = setpointings
        mytmp['ptgfile'] = ptgfile
        mytmp['integration'] = integration
        mytmp['direction'] = direction
        mytmp['mapsize'] = mapsize
        mytmp['antennalist'] = antennalist
        mytmp['hourangle'] = hourangle
        mytmp['totaltime'] = totaltime
        mytmp['tpnant'] = tpnant
        mytmp['tptime'] = tptime
        mytmp['pwv'] = pwv
        mytmp['image'] = image
        mytmp['imsize'] = imsize
        mytmp['imdirection'] = imdirection
        mytmp['cell'] = cell
        mytmp['niter'] = niter
        mytmp['threshold'] = threshold
        mytmp['graphics'] = graphics
        mytmp['verbose'] = verbose
        mytmp['overwrite'] = overwrite
	pathname='file:///'+os.environ.get('CASAPATH').split()[0]+'/xml/'
        trec = casac.utils().torecord(pathname+'simalma.xml')

        casalog.origin('simalma')
        if not trec.has_key('simalma') or not casac.utils().verify(mytmp, trec['simalma']) :
	    return False


	try :
          casalog.post('')
          casalog.post('##########################################')
          casalog.post('##### Begin Task: simalma           #####')
          casalog.post('')
          result = simalma(project, dryrun, skymodel, inbright, indirection, incell, incenter, inwidth, complist, compwidth, setpointings, ptgfile, integration, direction, mapsize, antennalist, hourangle, totaltime, tpnant, tptime, pwv, image, imsize, imdirection, cell, niter, threshold, graphics, verbose, overwrite)
          casalog.post('')
          casalog.post('##### End Task: simalma           #####')
          casalog.post('##########################################')


# saveinputs for individule engine has no use
# saveinputs should alos be removed from casa_in_py.py
#
#
#          saveinputs = myf['saveinputs']
#          saveinputs('simalma', 'simalma.last', myparams)
#
#
	except Exception, instance:
	  if(myf.has_key('__rethrow_casa_exceptions') and myf['__rethrow_casa_exceptions']) :
	     raise
	  else:
	     tname = 'simalma'
	     casalog.post('An error occurred running task %s: %s' % (tname,instance), 'ERROR')
	     pass

        gc.collect()
        return result
#
#
##
#    def paramgui(self, useGlobals=True):
#        """
#        Opens a parameter GUI for this task.  If useGlobals is true, then any relevant global parameter settings are used.
#        """
#        import paramgui
#
#        a=inspect.stack()
#        stacklevel=0
#        for k in range(len(a)):
#          if (string.find(a[k][1], 'ipython console') > 0) or (string.find(a[k][1], '<string>') >= 0):
#            stacklevel=k
#            break
#        myf = sys._getframe(stacklevel).f_globals
#
#        if useGlobals:
#            paramgui.setGlobals(myf)
#        else:
#            paramgui.setGlobals({})
#
#        paramgui.runTask('simalma', myf['_ip'])
#        paramgui.setGlobals({})
#
#
#
#
    def defaults(self, param=None):
        a=inspect.stack()
        stacklevel=0
        for k in range(len(a)):
          if (string.find(a[k][1], 'ipython console') > 0) or (string.find(a[k][1], '<string>') >= 0):
                stacklevel=k
                break
        myf=sys._getframe(stacklevel).f_globals
        a = odict()
        a['project']  = 'sim'
        a['dryrun']  = True
        a['skymodel']  = ''
        a['complist']  = ''
        a['setpointings']  = True
        a['antennalist']  = ['alma.cycle1.1.cfg', 'aca.cycle1.cfg']
        a['hourangle']  = 'transit'
        a['totaltime']  = ['20min', '1h']
        a['tpnant']  = 0
        a['pwv']  = 0.5
        a['image']  = True
        a['graphics']  = 'both'
        a['verbose']  = False
        a['overwrite']  = False

        a['skymodel'] = {
                    0:odict([{'notvalue':''}, {'inbright':''}, {'indirection':''}, {'incell':''}, {'incenter':''}, {'inwidth':''}])}
        a['complist'] = {
                    0:odict([{'notvalue':''}, {'compwidth':'8GHz'}])}
        a['setpointings'] = {
                    0:odict([{'value':True}, {'integration':'10s'}, {'direction':''}, {'mapsize':['', '']}]), 
                    1:odict([{'value':False}, {'ptgfile':'$project.ptg.txt'}, {'integration':'10s'}])}
        a['tpnant'] = {
                    0:odict([{'notvalue':0}, {'tptime':'0s'}])}
        a['image'] = {
                    0:odict([{'value':True}, {'imsize':0}, {'imdirection':''}, {'cell':''}, {'niter':0}, {'threshold':'0.1mJy'}]), 
                    1:{'value':False}}

### This function sets the default values but also will return the list of
### parameters or the default value of a given parameter
        if(param == None):
                myf['__set_default_parameters'](a)
        elif(param == 'paramkeys'):
                return a.keys()
        else:
	        if(a.has_key(param)):
		   #if(type(a[param]) == dict) :
		   #   return a[param][len(a[param])-1]['value']
	   	   #else :
		      return a[param]


#
#
    def check_params(self, param=None, value=None):
      a=inspect.stack() 
      stacklevel=0
      for k in range(len(a)):
        if (string.find(a[k][1], 'ipython console') > 0) or (string.find(a[k][1], '<string>') >= 0):
	    stacklevel=k
	    break
      myf=sys._getframe(stacklevel).f_globals

#      print 'param:', param, 'value:', value
      try :
         if str(type(value)) != "<type 'instance'>" :
            value0 = value
            value = myf['cu'].expandparam(param, value)
            matchtype = False
            if(type(value) == numpy.ndarray):
               if(type(value) == type(value0)):
                  myf[param] = value.tolist()
               else:
                  #print 'value:', value, 'value0:', value0
                  #print 'type(value):', type(value), 'type(value0):', type(value0)
                  myf[param] = value0
                  if type(value0) != list :
                     matchtype = True
            else :
               myf[param] = value
            value = myf['cu'].verifyparam({param:value})
            if matchtype:
               value = False
      except Exception, instance:
         #ignore the exception and just return it unchecked
         myf[param] = value
      return value

#
#
    def description(self, key='simalma', subkey=None):
        desc={'simalma': 'Simulation task for ALMA ',
               'project': 'root prefix for output file names',
               'dryrun': 'dryrun=True will only produce the informative report, not run simobserve/analyze',
               'skymodel': 'model image to observe',
               'inbright': 'scale surface brightness of brightest pixel e.g. "1.2Jy/pixel"',
               'indirection': 'set new direction e.g. "J2000 19h00m00 -40d00m00"',
               'incell': 'set new cell/pixel size e.g. "0.1arcsec"',
               'incenter': 'set new frequency of center channel e.g. "89GHz" (required even for 2D model)',
               'inwidth': 'set new channel width e.g. "10MHz" (required even for 2D model)',
               'complist': 'componentlist to observe',
               'compwidth': 'bandwidth of components',
               'setpointings': '',
               'ptgfile': 'list of pointing positions',
               'integration': 'integration (sampling) time',
               'direction': '"J2000 19h00m00 -40d00m00" or "" to center on model',
               'mapsize': 'angular size of map or "" to cover model',
               'antennalist': 'antenna position files of ALMA 12m and 7m arrays',
               'hourangle': 'hour angle of observation center e.g. -3:00:00, or "transit"',
               'totaltime': 'total time of observation; vector corresponding to antennalist',
               'tpnant': 'Number of total power antennas to use (0-4)',
               'tptime': 'total observation time for total power',
               'pwv': 'Precipitable Water Vapor in mm. 0 for noise-free simulation',
               'image': 'image simulated data',
               'imsize': 'output image size in pixels (x,y) or 0 to match model',
               'imdirection': 'set output image direction, (otherwise center on the model)',
               'cell': 'cell size with units or "" to equal model',
               'niter': 'maximum number of iterations (0 for dirty image)',
               'threshold': 'flux level (+units) to stop cleaning',
               'graphics': 'display graphics at each stage to [screen|file|both|none]',
               'verbose': '',
               'overwrite': 'overwrite files starting with $project',

              }

#
# Set subfields defaults if needed
#
        if(subkey == 'True'):
          desc['direction'] = 'center of map or "" to center on the model'
        if(subkey == 'False'):
          desc['integration'] = 'integration time (see below)'

        if(desc.has_key(key)) :
           return desc[key]

    def itsdefault(self, paramname) :
        a = {}
        a['project']  = 'sim'
        a['dryrun']  = True
        a['skymodel']  = ''
        a['inbright']  = ''
        a['indirection']  = ''
        a['incell']  = ''
        a['incenter']  = ''
        a['inwidth']  = ''
        a['complist']  = ''
        a['compwidth']  = '"8GHz"'
        a['setpointings']  = True
        a['ptgfile']  = '$project.ptg.txt'
        a['integration']  = '10s'
        a['direction']  = ['']
        a['mapsize']  = ['', '']
        a['antennalist']  = ['alma.cycle1.1.cfg', 'aca.cycle1.cfg']
        a['hourangle']  = 'transit'
        a['totaltime']  = ['20min', '1h']
        a['tpnant']  = 0
        a['tptime']  = '0s'
        a['pwv']  = 0.5
        a['image']  = True
        a['imsize']  = [128, 128]
        a['imdirection']  = ''
        a['cell']  = ''
        a['niter']  = 0
        a['threshold']  = '0.1mJy'
        a['graphics']  = 'both'
        a['verbose']  = False
        a['overwrite']  = False

        if a.has_key(paramname) :
	      return a[paramname]
simalma_pg = simalma_pg_()
